from typing import Union, List
from .common import ArincWord, reverse_label




class Encoder:
    def __init__(self) -> None:
        """Initialize an ARINC429 encoder with default values.

        Attributes:
            data (Union[int,float]): The encoded data value after processing
            label (int): 8-bit ARINC429 label (0-255)
            sdi (int): Source/Destination Identifier (0-3)
            ssm (int): Sign/Status Matrix (0-3)
            value (Union[int,float]): Raw input value before encoding
            encoding (str): Encoding type ("BNR", "BCD", or "DSC")
            msb (int): Most Significant Bit position (11-29)
            lsb (int): Least Significant Bit position (9-29)
            offset (Union[int,float]): BNR encoding offset value
            scale (Union[int,float]): BNR encoding scale factor
            word (int): Final 32-bit ARINC429 word
            b_arr (bytes): Byte array representation of the ARINC word
        """
        # General stuff
        self.data: Union[int, float] = 0
        self.label: int = 0
        self.sdi: int = 0
        self.ssm: int = 0
        self.value: Union[int, float] = 0
        self.encoding: str = ""
        self.msb: int = 29
        self.lsb: int = 11
        self.encodings: List[str] = []  # List of encodings used
        # BRN encoding
        self.offset: Union[int, float] = 0
        self.scale: Union[int, float] = 1
        # Output stuff
        self.word_val: int = 0
        self.b_arr_val: bytes = b"0"
        # we will use this if more than one encoding is used
        self.word_list: List[int] = []
        self.blist: List[bytes] = []
        # Byte list in int
        self.a429vals: List[ArincWord] = []

    def __repr__(self) -> str:
        return '\n'.join([str(word) for word in self.a429vals])

    def encode(
        self,
        value: Union[int, float] = 0,
        msb: int = 29,
        lsb: int = 11,
        label: int = 0,
        sdi: int = 0,
        ssm: int = 0,
        scale: float = 1,
        offset: float = 0,
        encoding: str = "",
    ) -> None:
        self.value = value
        self.label = label
        self.sdi = sdi
        self.ssm = ssm
        self.encoding = encoding
        self.msb = msb
        self.lsb = lsb
        self.offset = offset
        self.scale = scale
        # Check if the input is valid
        self._check_sdi()
        self._check_ssm()
        self._check_msb()
        self._check_lsb()

        if self.encoding == "BNR":
            self._encode_bnr()
        elif self.encoding == "BCD":
            self._encode_bcd()
        elif self.encoding == "DSC":
            self._encode_dsc()
        else:
            raise ValueError(f"Encoding {self.encoding} not supported")

        self._add_encoding(self.encoding)

    def _check_sdi(self):
        if not 0 <= self.sdi <= 0x03:
            raise ValueError("The SDI cannot be negative or bigger than 0x03")
        if self.lsb < 11 and self.sdi != 0:
            raise ValueError("SDI must be 0 if LSB is smaller than 11")

    def _check_ssm(self):
        if not 0 <= self.ssm <= 0x03:
            raise ValueError("The SSM cannot be negative or bigger than 0x03")

    def _check_msb(self,msb:Union[None,int]=None):
        if msb: #use the method somewhere else
            if not 11 <= msb <= 29:
                raise ValueError(
                    "The most significant bit cannot be bigger than 29 or smaller than 11")


        if not 11 <= self.msb <= 29:
            raise ValueError(
                "The most significant bit cannot be bigger than 29 or smaller than 11"
            )
    def _check_can_enc_dsc(self,msb)->None:
        if self.lsb < msb <self.msb:
            raise ValueError("DSC encoding can only be used out of the used range")
        

    def _check_lsb(self):
        if not 9 <= self.lsb <= 29:
            raise ValueError(
                "The least significant bit cannot be bigger than 29 or smaller than 9"
            )

    def add_dsc(self,value:int,msb:int):
        """
        Add a DSC encoded value to your A429 word
        """
        if not self.a429vals:
            raise ValueError("You need to encode a value first before adding DSC values. Use the encode method")
        self._add_encoding("DSC")
        self._check_msb(msb)
        self._check_can_enc_dsc(msb)

        byte1 = self.a429vals[-1].byte1
        byte2 = self.a429vals[-1].byte2
        byte3 = self.a429vals[-1].byte3
        byte4 = self.a429vals[-1].byte4

        byte4 &= 0x80 # Clear the parity bit

        if msb <17:
            byte2 = (byte2& ~(1 << 16- msb)) | (value<< 16- msb)
        elif msb <25:
            byte3 = (byte3& ~(1 << 24- msb)) | (value<< 24- msb)
        else:
            byte4 = (byte4& ~(1 << 29- msb)) | (value<< 29- msb)

        parity = self._get_parity(bytes([byte1, byte2, byte3, byte4]))
        if parity:  # If not, the parity is already set to zero
            byte4 |= 0x80

        self.a429vals[-1].byte2 = byte2
        self.a429vals[-1].byte3 = byte3
        self.a429vals[-1].byte4 = byte4
        self.a429vals[-1].encoding += " & DSC"
        self._build_word(byte4,byte3,byte2,byte1)
        
    @property
    def word(self) -> int:

        return self.word_val

    @property
    def bword(self) -> bytes:
        return self.a429vals[-1].get_bytes()

    def _can_bnr(self)->None:
        """
        Check if the value can be encoded in the BNR range 
        """

        nbits = int(self.data).bit_length()
        if nbits > self.msb - self.lsb:
            raise ValueError(
                f"Value {self.data} requires {nbits} bits. It cannot fit in the range {self.msb} to {self.lsb}"
            )


    def _encode_bnr(self):
        """
        Encode following the BNR schema

        data = (value - offset) / offset
        """
        self.data = (self.value - self.offset) / self.scale
        self._can_bnr()

        # Byte1 - label
        byte1 = self._reverse_label(self.label)
        mov = 2
        # Byte2 - SDI + some word stuff

        if self.lsb > 11:
            self.data = int(self.data) << (self.lsb - 11)

        elif self.lsb < 11:
            self.data = int(self.data) >> (11 - self.lsb)

        byte2 = self.sdi
        byte2 |= (int(self.data)) << mov
        byte2 &= 0xFF
        # Byte 3: Data
        byte3 = 0
        byte3 |= int(self.data) >> (mov + 4)
        byte3 &= 0xFF
        

        # Byte 4- Data + SSM + Parity
        byte4 = 0
        byte4 |= (int(self.data) >> (mov + 12)) & 0x3F
        byte4 |= self.ssm << 5

        parity = self._get_parity(bytes([byte1, byte2, byte3, byte4]))

        if parity:  # If not, the parity is already set to zero
            byte4 |= 0x80


        word = ArincWord(label=self.label,
                         byte1=byte1,
                         byte2=byte2,
                         byte3=byte3,
                         byte4=byte4,
                         encoding=self.encoding,
                         msb=self.msb,
                         lsb=self.lsb,
                         sdi=self.sdi,
                         ssm=self.ssm,
                         value=self.value,
                         offset=self.offset,
                         scale=self.scale,
                         data=self.data)
        self.a429vals.append(word)
        self._build_word(byte4,byte3,byte2,byte1)

    def _msb_mask(self, msb) -> int:
        masks = {
            # byte
            28: 0b1110111,
            27: 0b1110011,
            26: 0b1110001,
            25: 0b1110000,
            # Byte3
            24: 0b0111111,
            23: 0b0011111,
            22: 0b0001111,
            21: 0b0000111,
            20: 0b0000011,
            19: 0b0000001,
            18: 0b0000000,
            # Byte2
            17: 0b0111111,
            16: 0b0011111,
            15: 0b0000111,
            14: 0b0000011,
            13: 0b0000001,
            12: 0b0000001,
            11: 0b0000000,
        }

        return masks[msb]

    def reset(self):
        """
        Reset the encoder to the initial state
        """
        self.data = 0
        self.label = 0
        self.sdi = 0
        self.ssm = 0
        self.value = 0
        self.encoding = ""
        self.msb = 29
        self.lsb = 11
        self.encodings = []
        self.a429vals = []

    def _encode_bcd(self):
        """
        BCD encoding for arinc429 data
        """
        mov = 2  # We dont care about MSB or LSB for BCD
        if self.value < 0:
            raise ValueError(
                "BCD encoding does not support negative values. Use BNR encoding instead."
            )

        self.data = self.value
        if self.value > 79999:  # Cant encode antyhing bigger than this
            self.data = self.data // 10
        # Encode data for BCD
        iterval = int(self.data)
        i = 0
        encVal = 0
        while iterval > 0:
            encVal |= (iterval % 10) << (4 * i)
            iterval //= 10
            i += 1
        self.data = encVal
        # Normal encoding process
        # Byte 1
        byte1 = self._reverse_label(self.label)

        # Byte 2
        byte2 = self.sdi
        byte2 |= (int(self.data) & 0x3F) << mov
        byte2 &= 0xFF

        # Byte 3: Data
        byte3 = 0
        byte3 |= int(self.data) >> (mov + 4)
        byte3 &= 0xFF
        # Byte 4- Data + SSM + Parity
        byte4 = 0
        byte4 |= (int(self.data) >> (mov + 12)) & 0x3F
        byte4 |= self.ssm << 5

        parity = self._get_parity(bytes([byte1, byte2, byte3, byte4]))

        if parity:  # If not, the parity is already set to zero
            byte4 |= 0x80


        # self.word_list.append(self.word_val)
        # self.blist.append(self.b_arr_val)

        word = ArincWord(label=self.label,
                         byte1=byte1,
                         byte2=byte2,
                         byte3=byte3,
                         byte4=byte4,
                         encoding=self.encoding,
                         msb=None,
                         lsb=None,
                         sdi=self.sdi,
                         ssm=self.ssm,
                         value=self.value,
                         offset=None,
                         scale=None,
                         data=self.data)
        self.a429vals.append(word)
        self._build_word(byte4,byte3,byte2,byte1)


    def _add_encoding(self, encoding: str) -> None:
        """
        Add an encoding to the list of previous encodings
        """
        # Making sure that only one BNR or BCD encoding is present

        if "BNR" in self.encodings or "BCD" in self.encodings:
            raise ValueError("Only one BNR or BCD encoding is allowed per word")

        self.encodings.append(encoding)

    def _encode_dsc(self):
        """
        DSC encoding for arinc429 data
        """
        mov = 2  # We dont care about MSB or LSB for BCD
        if not ((self.value == 0) or (self.value == 1)):
            raise ValueError(
                "DSC encoding does not support values other than 0 or 1. Use BCD/BNR encoding instead."
            )

        self.data = self.value

        # Encode data for DSC
        # Byte 1
        byte1 = self._reverse_label(self.label)

        # Byte 2
        byte2 = self.sdi
        byte2 |= (int(self.data) & 0x3F) << mov
        byte2 &= 0xFF

        # Byte 3: Data
        byte3 = 0
        byte3 |= int(self.data) >> (mov + 4)
        byte3 &= 0xFF
        # Byte 4- Data + SSM + Parity
        byte4 = 0
        byte4 |= (int(self.data) >> (mov + 12)) & 0x3F
        byte4 |= self.ssm << 5

        parity = self._get_parity(bytes([byte1, byte2, byte3, byte4]))

        if parity:  # If not, the parity is already set to zero
            byte4 |= 0x80

        word = ArincWord(label=self.label,
                         byte1=byte1,
                         byte2=byte2,
                         byte3=byte3,
                         byte4=byte4,
                         encoding=self.encoding,
                         msb=None,
                         lsb=None,
                         sdi=self.sdi,
                         ssm=self.ssm,
                         value=self.value,
                         offset=None,
                         scale=None,
                         data=self.data)
        self.a429vals.append(word)
        self._build_word(byte4,byte3,byte2,byte1)


    def _get_parity(self, b_data: bytes) -> bool:
        """
        Computes the odd parity for the entire 32-bit ARINC429 word.
        Returns True if parity bit should be 1, False if it should be 0
        to maintain odd parity.
        """
        # Count all 1 bits in the entire word (excluding the parity bit)
        num_ones = 0
        for byte in b_data:
            # For the last byte, mask out the parity bit (MSB)
            if byte == b_data[-1]:
                byte &= 0x7F
            num_ones += bin(byte).count("1")

        return num_ones % 2 == 0

    def _reverse_label(self, label: int) -> int:
        self.rlabel = reverse_label(label)
        return self.rlabel

    @property
    def data_val(self) -> Union[int, float]:
        """
        Return the value of the processed value to be encoded
        """
        return self.data

    def _build_word(self,byte4,byte3,byte2,byte1):
        """
        Build the arinc429 32 bite word form the invidivual bytes
        """
        self.word_val = (byte4<<24) | (byte3<<16) | (byte2<<8) | byte1




from arinc429.common import ArincWord,reverse_label



def test_ArincWord():
    w = ArincWord(byte4=0x10,byte3=0x20,byte2=0x30,byte1=0x40)
    assert w.word == 0x10203040
    assert w.get_bytes() == bytes([0x10, 0x20, 0x30, 0x40])

def test_reverse_label():
    label = 0o101
    assert reverse_label(label) == 0b10000010



from .common import ArincWord, reverse_label
import struct
from typing import List,Union


class Decoder:
    def __init__(self):
        self.dword:ArincWord = ArincWord()

    def decode(self, 
               word: Union[int, List[int],bytes],
               encoding: str,
               msb: int = 29,
               lsb: int = 11,
               scale: float = 1,
               offset: float = 0,
               sdi: int = 0,
               ssm: int = 0
               )->ArincWord:
        """
        Decode main function.


        Parameters
        ----------
        word : Union[int, List[int],bytes]
            The ARINC word to decode. It can be an integer, a list of integers or a bytes object.
        encoding : str
        msb : int, Most significant bit
        lsb : int, Least significant bit
        scale : float
        offset : float
        sdi : int
        ssm : int
        """
        if isinstance(word, int):
            word = struct.pack("<I",word)
        elif isinstance(word, list):
            word = bytes(word)
        elif isinstance(word, bytes):
            pass


        self.dword.encoding = encoding
        self.dword.byte1 = word[3]# Label
        self.dword.byte2 = word[2]
        self.dword.byte3 = word[1]
        self.dword.byte4 = word[0] 
        self.dword.msb = msb
        self.dword.lsb = lsb
        self.dword.scale = scale
        self.dword.offset = offset
        self.dword.sdi = sdi
        self.dword.ssm = ssm

        self._decode_label()
        self._decode_sdi_ssm()

        if encoding == "BCD":
            return self._decode_bcd()
        elif encoding == "BNR":
            return self._decode_bnr()
        elif encoding == "DSC":
            return self._decode_dsc()
        else:
            raise ValueError("Unknown encoding")

    def _decode_label(self):
        self.dword.label = reverse_label(self.dword.byte1)

    def _decode_bcd(self)->ArincWord:
        self.dword.data = ((self.dword.byte4 & 0x1F)<<14) | ((self.dword.byte3)<<6) | ((self.dword.byte2 & 0xFC)>>2)
        self.dword.data = self.dword.data >> (self.dword.lsb -11) if self.dword.lsb >= 11 else self.dword.data << (11-self.dword.lsb)
        return self.dword

    def _decode_sdi_ssm(self):

        if self.dword.lsb<11:
            self.dword.sdi = 0
        else:
            self.dword.sdi = self.dword.byte2 & 0x03

        if self.dword.msb>29:
            self.dword.ssm = 0
        else:
            self.dword.ssm = (self.dword.byte4 & 0x60) >> 5

    def _decode_bnr(self)->ArincWord:
        """
         I need to get from the bytes, the ssm, the sdi, the parity and the value

        edata = (value - offset)/scale

        TODO: Add the MSB and LSB
        """

        self.dword.data = ((self.dword.byte4 & 0x1F)<<14) | ((self.dword.byte3)<<6) | ((self.dword.byte2 & 0xFC)>>2)
        self.dword.data = self.dword.data >> (self.dword.lsb -11) if self.dword.lsb >= 11 else self.dword.data << (11-self.dword.lsb)
        

        self.dword.value = float(self.dword.data) * self.dword.scale + self.dword.offset
        return self.dword

    def _decode_dsc(self)->ArincWord:
        pass
        return self.dword



from arinc429 import Encoder
import pytest

def test_enc_bnr_zero():
    a429 = Encoder()
    a429.encode(encoding="BNR")
    assert(a429.bword== b"\x80\x00\x00\x00")

def test_enc_brn_std():
    a429 = Encoder()
    det={
            "label":0o205,
            "value":100,
            "ssm": 0x03,
            "sdi":0,
            "encoding":"BNR"
            }
    a429.encode(**det)
    assert(a429.bword== b"\xe0\x01\x90\xa1")
    assert(a429.word== 0xe00190a1)

def test_enc_brn_ssm1():
    a429 = Encoder()
    det={
            "label":0o206,
            "value":105,
            "ssm": 0x01,
            "sdi":0,
            "encoding":"BNR"
            }
    a429.encode(**det)
    assert(a429.bword== b"\xa0\x01\xa4\x61")
    assert(a429.word== 0xa001a461)

def test_enc_brn_msb_lsb():
    a429 = Encoder()
    det={
            "label":0o321,
            "value":50,
            "ssm": 0x03,
            "sdi":0,
            "msb": 25,
            "lsb": 15,
            "encoding":"BNR"
            }
    a429.encode(**det)
    assert(a429.bword== b"\x60\x0c\x80\x8b")
    assert(a429.word== 0x600c808b)

def test_enc_brn_msb_lsb2():
    a429 = Encoder()
    det={
            "label":0o333,
            "value":127.7,
            "ssm": 0x03,
            "sdi":0,
            "msb": 23,
            "lsb": 16,
            "scale":1,
            "encoding":"BNR"
            }
    a429.encode(**det)
    assert(a429.bword== b"\x60\x3F\x80\xdb")
    assert(a429.word== 0x603f80db)

def test_enc_brn_msb_not_enough_bits():
    a429 = Encoder()
    det={
            "label":0o333,
            "value":127.7,
            "ssm": 0x03,
            "sdi":0,
            "msb": 21,
            "lsb": 16,
            "encoding":"BNR"
            }
    with pytest.raises(ValueError):
        a429.encode(**det)
# TODO: Implement test_enc_bnr_msb_lsb_more_info_no_sdi
# Note: this encodes soemthing like longitude  by using the SDI for the extra precision
@pytest.mark.skip(reason="Not implemented")
def test_enc_bnr_msb_lsb_more_info_no_sdi():
    a429 = Encoder()
    det={
            "label":0o110,
            "value":42.2431231,
            "ssm": 0x03,
            "sdi":0,
            "msb": 29,
            "lsb": 9,
            "scale":1,
            "offset":0,
            "encoding":"BNR"
            }
    a429.encode(**det)
    assert(a429.bword== b"\x60\x3F\x80\xdb")
    assert(a429.word== 0x603f80db)




from arinc429 import Encoder


def test_enc_bcd_zero():
    a429 = Encoder()
    a429.encode(encoding="BCD")
    assert a429.bword== b"\x80\x00\x00\x00"
    assert a429.word== 0x80000000

def test_enc_bcd_case1():
    a429 = Encoder()
    det={
            "label":0o205,
            "value":78501,
            "ssm": 0x03,
            "sdi":0,
            "encoding":"BCD"
            }
    a429.encode(**det)
    assert(a429.bword== b"\xfe\x14\x04\xa1")
    assert(a429.word== 0xfe1404a1)

def test_enc_bcd_case2():
    a429 = Encoder()
    det={
            "label":0o206,
            "value":80001,
            "ssm": 0x03,
            "sdi":0,
            "encoding":"BCD"
            }
    a429.encode(**det)
    assert a429.bword== b"\xe2\x00\x00\x61"
    assert a429.word== 0xe2000061

def test_enc_bcd_case3():
    a429 = Encoder()
    det={
            "label":0o206,
            "value":80001,
            "ssm": 0x03,
            "sdi":1,
            "encoding":"BCD"
            }
    a429.encode(**det)
    assert a429.bword== b"\x62\x00\x01\x61"
    assert a429.word== 0x62000161



from dataclasses import dataclass
from typing import Union

@dataclass
class ArincWord:
    """
    Container for an ARINC 429 word.

    """
    label: int = 0
    byte1: int = 0
    byte2: int = 0
    byte3: int = 0
    byte4: int = 0
    encoding: str = ""
    msb: int  = 0
    lsb: int  = 0
    sdi: int = 0
    ssm: int = 0
    value: Union[int, float] = 0
    offset: Union[int, float] = 0
    scale:  float = 1
    data: Union[int, float] = 0
    max: Union[int,float,None] = None
    min: Union[int,float,None] = None

    def __repr__(self):
                return f"ArincWord(0x{self.byte1:02x},0x{self.byte2:02x},0x{self.byte3:02x},0x{self.byte4:02x})"
    @property
    def word(self):
        """
        Returns the uint32 arinc word from bytes
        """
        return int.from_bytes(bytes([self.byte1,self.byte2,self.byte3,self.byte4]),byteorder="little",signed=False)

    def get_bytes(self)->bytes:
        """
        Retruns the bytes of the arinc Word
        """
        return bytes([self.byte4,self.byte3,self.byte2,self.byte1])

    @property
    def parity(self):
        """
        Returns the parity of the words
        """
        return self.byte4 & 0x80

    def visualize(self)->str:
        """Returns a string visualization of the ARINC word bits in markdown table format

        Like this :
        |32|31|30|29|28|27|26|25|24|23|22|21|20|19|18|17|16|15|14|13|12|11|10|9 |8 |7 |6 |5 |4 |3 |2 |1 |
        |--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|--|
        |0 |1 |1 |0 |0 |0 |0 |0 |0 |0 |1 |1 |1 |1 |1 |1 |1 |0 |0 |0 |0 |0 |0 |0 |1 |1 |0 |1 |1 |0 |1 |1 |

        This is really useful for debugging purposes or to visualize the bits of an ARINC word

        """
        bits = []
        for byte in [self.byte4, self.byte3, self.byte2, self.byte1]:
            bits.extend([str((byte >> i) & 1) for i in range(7,-1,-1)])
        
        header = "|" + "|".join(f"{i:^2d}" for i in range(32,0,-1)) + "|"
        separator = "|" + "|".join("--" for _ in range(32)) + "|"
        bits_str = "|" + "|".join(f"{b:^2}" for b in bits) + "|"
        return f"{header}\n{separator}\n{bits_str}"


def reverse_label(label: int) -> int:
    """
    Reverses the bits of an 8-bit unsigned integer using bitwise operations.
    """
    if not 0 <= label <= 255:
        raise ValueError("Input must be an 8-bit unsigned integer (0 <= n <= 255).")

    label = ((label & 0b11110000) >> 4) | ((label & 0b00001111) << 4)
    label = ((label & 0b11001100) >> 2) | ((label & 0b00110011) << 2)
    label = ((label & 0b10101010) >> 1) | ((label & 0b01010101) << 1)

    return label



from arinc429 import Decoder, ArincWord
import pytest


def test_dec_bnr_zero():
    a429 = Decoder()
    word = a429.decode([0,0,0,0], 
                encoding="BNR",
                )
    assert(word.value == 0)
def test_dec_bnr_std():
    a429 = Decoder()
    word = a429.decode(
           b"\xe0\x01\x90\xa1",
                encoding="BNR",
                )
    assert(word.label== 0o205)
    assert(word.ssm== 0x03)
    assert(word.sdi== 0)
    assert(word.value == 100)

def test_dec_bnr_ssm1():
    a429 = Decoder()
    word = a429.decode(
                b"\xa0\x01\xa4\x61",
                encoding="BNR",
                )
    assert(word.label== 0o206)
    assert(word.ssm== 0x01)
    assert(word.sdi== 0)
    assert(word.value == 105)

def test_dec_brn_msb_lsb():
    a429 = Decoder()
    word = a429.decode(
            b"\x60\x0c\x80\x8b",
                encoding="BNR",
                lsb = 15
                )
    assert(word.label== 0o321)
    assert(word.ssm== 0x03)
    assert(word.sdi== 0)
    assert(word.value == 50)

def test_dec_brn_msb_lsb2():
    a429 = Decoder()
    word = a429.decode(
            b"\x60\x3F\x80\xdb",
                encoding="BNR",
                lsb =16 ,
                
                )
    assert(word.label== 0o333)
    assert(word.ssm== 0x03)
    assert(word.sdi== 0)
    assert(word.value == 127)


